module Result = struct
  type error =
    { stack : (string * TextInput.Position.t) list;
      last_position : TextInput.Position.t
    }
    
  type 'value output =
    { value : 'value;
      interval : TextInput.TextInterval.t;
      input : TextInput.t;
      latest_error : error option
    }

  let compare_errors err1 err2 =
    TextInput.Position.compare err1.last_position err2.last_position

  let latest err_opt1 err_opt2 =
    match err_opt1, err_opt2 with
    | Some err1, Some err2 when compare_errors err1 err2 < 0 -> err_opt2
    | Some err1, _ -> err_opt1
    | _ -> err_opt2

  type 'value t =
    | Output of 'value output
    | Error of error

  let ok ~value ~interval ~input ~latest_error =
    Output { value; interval; input; latest_error }

  let error ~msg ~position =
    Error { stack = [(msg,position)]; last_position = position }


  let add_error err1 = function
    | Error err2 when compare_errors err1 err2 > 0 -> Error err1
    | Error err2 -> Error err2
    | Output out ->
      Output { out with
               latest_error = latest (Some err1) out.latest_error
             }

  let bind result fct =
    match result with
    | Output value -> fct value
    | Error err -> Error err
  let (>>=) = bind

  let map result ~f =
    match result with
    | Output out -> Output { out with value = f out.value }
    | Error _ -> result

  let is_successful = function
    | Error err -> false
    | Output out -> true
      
  let get_value = function
    | Error err -> None
    | Output out -> Some out.value

  let get_interval = function
    | Error err -> None
    | Output out -> Some out.interval

  let get_input_left = function
    | Error err -> None
    | Output out -> Some out.input

  let get_stack = function
    | Error err ->
      Some (err.stack |> List.rev )
    | Output { latest_error = Some err } -> Some err.stack
    | Output _ -> None

  let get_error_position = function
    | Error err -> Some err.last_position
    | Output out -> None

end

let (--) = TextInput.TextInterval.merge

type 'value result = 'value Result.t 
type 'value reader = TextInput.t -> 'value result

let run reader input = reader input


let char : char reader =
  fun input ->
    match TextInput.observe input with
    | None ->
      Result.error
        ~msg:"End of input"
        ~position:(TextInput.get_current_position input)
    | Some ((char,position), stream) ->
      Result.ok
        ~value:char
        ~interval:(TextInput.TextInterval.of_position position)
        ~input:stream
        ~latest_error:None

let return (value:'value) : 'value reader =
  fun input ->
    let position = TextInput.get_current_position input in
    Result.ok
        ~value
        ~interval:(TextInput.TextInterval.of_position position)
        ~input
        ~latest_error:None

let or_else reader1 reader2 input =
  match reader1 input with
  | Result.Output out as ok -> ok
  | Result.Error err ->
    reader2 input 
    |> Result.add_error err 

let apply : ('arg -> 'im) reader -> 'arg reader -> 'im reader =
  fun fun_reader arg_reader input ->
  let open Result in
  fun_reader input >>= fun funct ->
  arg_reader funct.input >>= fun arg ->
  let latest_error =
    Result.latest funct.latest_error arg.latest_error
  in
  ok
    ~value:(funct.value arg.value)
    ~interval:(funct.interval -- arg.interval)
    ~input:arg.input
    ~latest_error


let map (arg_reader: 'arg reader) ~(f:'arg -> 'im) : 'im reader =
  fun input ->
    let open Result in
    arg_reader input >>= fun arg ->
    Result.Output { arg with value = f arg.value }

let bind
    (arg_reader: 'arg reader)
    ~f:(make_reader: 'arg -> 'im reader)
  : 'im reader =
  fun input ->
    let open Result in
    arg_reader input >>= fun arg ->
    (make_reader arg.value) arg.input >>= fun result ->
    Output { result with interval = arg.interval -- result.interval }

let fail : string -> 'any reader =
  fun msg input ->
    Result.error ~msg ~position:(TextInput.get_current_position input)

let rec fix : ('value reader -> 'value reader) -> 'value reader =
  fun make_reader input ->
    run (make_reader (fix make_reader)) input


let rule : name:string -> 'value reader -> 'value reader =
  fun ~name reader input ->
    let open Result in
    let position = TextInput.get_current_position input in
    let push err = { err with stack = (name,position) :: err.stack } in
    match reader input with
    | Output out ->
      Output { out with
               latest_error = Option.(out.latest_error >|= push)
             }
    | Error err -> Error (push err) 


let (||) = or_else
let (<*>) = apply
let ( *> ) fct reader = map reader ~f:fct

let (>>=) arg fct = bind arg ~f:fct


let sequence (reader: 'elt reader) : 'elt list reader =
  fix @@ fun seq -> 
  (fun elt list -> elt :: list) *> reader <*> seq
  || return []

let non_empty_sequence (reader: 'elt reader) : 'elt list reader =
  (fun elt list -> elt :: list) *> reader <*> sequence reader

let satisfying ?(msg="condition not satisfied")
    (condition: 'value -> bool)
    (reader: 'value reader)
  : 'value reader =
  reader >>= fun value ->
  if condition value then return value
  else fail msg

let (@?) condition reader = satisfying ~msg:"condition not satisfied" condition reader


let list_of_string str =
  Array.init (String.length str) (String.get str)
  |> Array.to_list

let string_of_list list =
  list
  |> List.map (fun c -> String.make 1 c)
  |> String.concat ""

module Char = struct
  
  let figure : char reader =
    satisfying
      ~msg:"not a figure"
      CharClasses.is_figure
      char

  let lowercase : char reader =
    satisfying
      ~msg:"not in lowercase"
      CharClasses.is_lowercase
      char
  let uppercase : char reader =
    satisfying
      ~msg:"not in uppercase"
      CharClasses.is_uppercase
      char
  let letter : char reader =
    satisfying
      ~msg:"not a letter"
      CharClasses.is_letter
      char
  let alphanumeric : char reader =
    satisfying
      ~msg:"not an alphanumeric character"
      CharClasses.is_alphanumeric
      char

  let one_char_of string : char reader =
    satisfying
      ~msg:(Printf.sprintf "expecting one of %s" string)
      (CharClasses.is_in (list_of_string string))
      char
      
  let this c : char reader =
    satisfying
      ~msg:(Printf.sprintf "expecting character %c" c)
      ((=) c)
      char
end

module Int = struct

  let convert_to_int o x leading figures =
    (o::x::leading::figures) |> string_of_list |> int_of_string

  let hexadecimal_figure =
    Char.figure
    || CharClasses.is_between 'a' 'f' @? char
    || CharClasses.is_between 'A' 'F' @? char

  let octal_figure =
    CharClasses.is_between '0' '7' @? char

  let binary_figure =
    CharClasses.is_between '0' '1' @? char

  let is_not_zero c = c <> '0'

  let pos_base ~base_symbols ~base_figure =
    convert_to_int
    *> Char.this '0'
    <*> Char.one_char_of base_symbols
    <*> is_not_zero @? base_figure
    <*> sequence ( base_figure || Char.this '_')

  let pos_hexadecimal =
    pos_base ~base_symbols:"xX" ~base_figure:hexadecimal_figure
  let pos_octal =
    pos_base ~base_symbols:"oO" ~base_figure:octal_figure
  let pos_binary =
    pos_base ~base_symbols:"bB" ~base_figure:binary_figure

  let pos_decimal =
    (fun leading figures ->
       leading::figures |> string_of_list |> int_of_string
    )
    *> (is_not_zero @? Char.figure)
    <*> sequence (Char.figure || Char.this '_')

  let positive_int =
    pos_hexadecimal
    || pos_octal
    || pos_binary
    || pos_decimal

  let negative_int =
    (fun _ n -> -n) *> Char.this '-' <*> positive_int

  let int =
    positive_int || negative_int

end


let int = Int.int


module String = struct

  let this string : string reader =
    let char_readers = 
      string
      |> list_of_string
      |> List.map Char.this
    in
    List.fold_right
      (fun char suffix -> (fun c list -> c :: list) *> char <*> suffix)
      char_readers
      (return [])
    |> map ~f:string_of_list

end

let s : string -> string reader = String.this


let space = CharClasses.is_white @? char
let skip_spaces = ignore *> sequence space
let spaces = non_empty_sequence space 


let (<+>) fct arg =
  (fun f spaces a -> f a) *> fct <*> skip_spaces <*> arg

let pair left right =
  (fun lpar left comma right rpar -> (left,right)) *>
  s "(" <+> left <+> s "," <+> right <+> s ")"

let triple left middle right =
  (fun lpar left comma1 middle comma2 right rpar -> (left, middle, right))
  *> s "(" <+> left <+> s "," <+> middle <+> s "," <+> right <+> s ")"


