type 'value reader
type 'value result

val run : 'value reader -> TextInput.t -> 'value result

module Result : sig
  val is_successful : 'value result -> bool 
  
  (** The valuer read by the parser (on success) *)
  val get_value : 'value result -> 'value option

  (** The interval of characters read by the parser (on success) *)
  val get_interval : 'value result -> TextInput.TextInterval.t option

  (** The text not read by the parser (on success) *)
  val get_input_left : 'value result -> TextInput.t option

  (** On error, gives the list of rules of the grammar used to get to
      the point when the error was detected. For each rule, gives the
      name of the rule, and the point from which the rule starts to
      apply. *)
  val get_stack : 'value result -> (string * TextInput.Position.t) list option

  (** On error, gives the position where the error occurred *)
  val get_error_position : 'value result -> TextInput.Position.t option
end


val return : 'value -> 'value reader

val apply : ('arg -> 'im) reader -> 'arg reader -> 'im reader
val (<*>) :  ('arg -> 'im) reader -> 'arg reader -> 'im reader
val (<+>) :  ('arg -> 'im) reader -> 'arg reader -> 'im reader
    
val map : 'arg reader -> f:('arg -> 'im) -> 'im reader
val ( *> ) : ('arg -> 'im)  -> 'arg reader -> 'im reader

val bind : 'arg reader -> f:('arg -> 'im reader) -> 'im reader
val (>>=) : 'arg reader -> ('arg -> 'im reader) -> 'im reader

val or_else : 'value reader -> 'value reader -> 'value reader
val (||) : 'value reader -> 'value reader -> 'value reader

val fail : string -> 'any reader

val rule : name:string -> 'value reader -> 'value reader

val fix : ('value reader -> 'value reader) -> 'value reader

val satisfying :
  ?msg:string -> ('value -> bool) -> 'value reader -> 'value reader
val (@?) : ('value -> bool) -> 'value reader -> 'value reader



val sequence : 'value reader -> 'value list reader
val non_empty_sequence : 'value reader -> 'value list reader

val char : char reader

val space : char reader
val skip_spaces : unit reader
val spaces : char list reader 

val pair : 'left reader -> 'right reader -> ('left * 'right) reader
val triple :
  'left reader ->
  'middle reader ->
  'right reader ->
  ('left * 'middle * 'right) reader
    
module Char : sig
  val figure : char reader
  val lowercase : char reader
  val uppercase : char reader
  val letter : char reader
  val alphanumeric : char reader
  val one_char_of : string -> char reader
  val this : char -> char reader
end

module Int : sig
  val positive_int : int reader
  val negative_int : int reader
end

val int : int reader

module String : sig
  val this : string -> string reader
end

val s : string -> string reader 
