type 'value reader = TextInput.t -> ('value * TextInput.t) option


let or_else (reader1: 'value reader) (reader2: 'value reader) : 'value reader =
  fun input ->
    match reader1 input with
    | Some result -> Some result
    | None -> reader2 input

let (||) = or_else

let fail : 'any reader = fun input -> None


let apply (fct_reader: ('arg -> 'im) reader)  (arg_reader: 'arg reader)
  : 'im reader =
  fun input ->
    let open Option in
    fct_reader input >>= fun (fct,remaining_input) ->
    arg_reader remaining_input >>= fun (arg,final_input) ->
    Some (fct arg, final_input)

let (<*>) = apply

let return : 'value -> 'value reader =
  fun value input -> Some (value,input)

let ( *> ) : ('arg -> 'im) -> 'arg reader -> 'im reader
  = fun fct reader -> return fct <*> reader


let char : char reader = fun input ->
  match TextInput.observe input with
  | None -> None
  | Some ((char,position),remaining_input) -> Some (char,remaining_input)



let satisfying ~condition (reader: 'value reader) : 'value reader = fun input ->
  match reader input with
  | Some (result,remaining) as output when condition result -> output
  | Some (bad,_) -> None
  | None -> None


let (@?) condition reader = satisfying ~condition reader

let this_char : char -> char reader =
  fun c -> char |> satisfying ~condition:(fun c' -> c = c')

let char_satisfying : (char -> bool) -> char reader =
  fun char_set -> char |> satisfying ~condition:char_set


let rec fix (make_reader : 'value reader -> 'value reader) : 'value reader =
  fun input ->
    make_reader (fix make_reader) input

let sequence : 'value reader -> 'value list reader =
  fun read_value ->
    fix @@ fun read_sequence -> 
    (fun value seq -> value :: seq) *> read_value <*> read_sequence
    || return []

let non_empty_sequence read_value =
  (fun value list -> value :: list) *> read_value <*> sequence read_value

let is_figure c = '0' <= c && c <= '9'
let is_non_zero_figure c = '1' <= c && c <= '9'

let make_int first_figure figures =
  first_figure :: figures
  |> List.map (fun c -> int_of_char c - int_of_char '0')
  |> List.fold_left (fun n figure -> 10 * n + figure) 0
let make_zero _ = 0

let int : int reader =
  make_int *> (is_non_zero_figure @? char) <*> sequence (is_figure @? char)
  || make_zero *> this_char '0'

let _ = int (TextInput.of_string "123abc")


let rec list_of_string ?(offset=0) string =
  if offset >= String.length string then []
  else String.get string offset :: list_of_string ~offset:(offset+1) string

let string_of_list chars =
  chars
  |> List.map (fun c -> String.make 1 c)
  |> String.concat ""


let this_string string : string reader =
  let (++) read_char read_string =
    (fun c list -> c::list) *> read_char <*> read_string
  in
  let char_readers =
    string
    |> list_of_string
    |> List.map (fun c -> this_char c)
  in
  return []
  |> List.fold_right (++) char_readers
  |> ( *> ) string_of_list

(* [this_string] is a bit long and will obfuscate the grammar,
   it is better to have a shot name for it. *)
let s = this_string


let is_space : char -> bool = String.contains " \t\r\n"

let one_space : char reader = is_space @? char
let skip_spaces : char list reader = sequence one_space
let spaces : char list reader = non_empty_sequence one_space


let (<+>) reader1 reader2 =
  (fun fct blanks arg -> fct arg) *> reader1 <*> skip_spaces <*> reader2


let pair left_reader right_reader : ('left * 'right) reader =
  let make_pair lparen left comma right rparen = (left,right) in
  make_pair *> s "(" <+> left_reader <+> s "," <+> right_reader <+> s ")"

let triple left_reader middle_reader right_reader
    : ('left * 'middle * 'right) reader =
  let make_triple lparen left comma1 middle comm2 right rparen =
    (left,middle,right)
  in
  make_triple *> s "("
    <+> left_reader <+> s ","
    <+> middle_reader <+> s ","
    <+> right_reader
    <+> s ")"



type 'elt bin_tree =
  | Leaf
  | Node of 'elt bin_tree * 'elt * 'elt bin_tree
              

let make_leaf leaf = Leaf
let make_node node (left,root,right) =
  Node (left,root,right)

let bin_tree (elt: 'elt reader) : 'elt bin_tree reader =
  fix @@ fun tree -> 
  make_node *> s "Node" <+> triple tree elt tree
  || make_leaf *> s "Leaf"  


let _ =
  bin_tree int
    (TextInput.of_string "Node ( Node( Leaf, 1 ,Leaf), 3, Leaf) blabla")
