let list_of_string string =
  Array.init (String.length string) (String.get string)
  |> Array.to_list

let string_of_list list =
  list
  |> List.map (String.make 1)
  |> String.concat ""


module Position = struct

  type t =
    { description : string;
      column : int;
      line : int;
    }

  let create description =
    { description;
      column = 1;
      line = 1;
    }

  let incr_by ~char pos =
    if char = '\n' then
      { pos with line = pos.line + 1; column = 1 }
    else
      { pos with column = pos.column + 1 }

  let incr_by_string ~string pos =
    string
    |> list_of_string
    |> List.fold_left (fun pos char -> incr_by ~char pos) pos


  let compare pos1 pos2 =
    if pos1.line <> pos2.line then pos1.line - pos2.line
    else if pos1.column <> pos2.column then pos1.column - pos2.column
    else compare pos1.description pos2.description

  let min pos1 pos2 = if compare pos1 pos2 < 0 then pos1 else pos2
  let max pos1 pos2 = if compare pos1 pos2 < 0 then pos2 else pos1

  let format fmt pos =
    Format.fprintf fmt "@[%s:%d:%d@]" pos.description pos.column pos.line

  let long_format fmt pos =
    Format.fprintf fmt "@[From %s, line %d, column %d@]"
      pos.description pos.column pos.line
end


module TextInterval = struct
  type t = Position.t * Position.t

  open Position

  let of_position pos = (pos, pos)

  let merge (start1,stop1) (start2,stop2) =
    (min start1 start2, max stop1 stop2)

  let ( -- ) interval_opt1 interval_opt2 =
    match interval_opt1, interval_opt2 with
    | None, _ -> interval_opt2
    | _, None -> interval_opt1
    | Some inter1, Some inter2 ->
      Some (merge inter1 inter2)

  let format fmt (start,stop) =
    Format.fprintf fmt "@[%s:%d:%d--%d:%d@]"
      start.description start.line start.column
      stop.line stop.column
end


type t =
  { stream : CharStream.t;
    position : Position.t;
  }

let get_current_position { position } = position 

let get_current_char { stream; position } =
  CharStream.get_current_char stream
  |> Option.map ~f:(fun c -> (c,position))

let move_forward { stream; position } =
  let open Option in
  CharStream.observe stream >|= fun (char,stream) ->
  let position = Position.incr_by ~char position in
  { stream; position }


let observe { stream; position } =
  let open Option in
  CharStream.observe stream >|= fun (char,stream) ->
  let position = Position.incr_by ~char position in
  ( (char,position), { stream; position } )


let of_string ?(offset=0) ?(name="string") string =
  { stream = CharStream.of_string ~offset string;
    position = Position.create name;
  }

let of_list ?(name="list") list =
  { stream = CharStream.of_list list;
    position = Position.create name
  }

let of_file ~filename =
  { stream = CharStream.of_file ~filename;
    position = Position.create filename;
  }

let of_in_channel ?(name="in_channel") in_channel =
  { stream = CharStream.of_in_channel in_channel;
    position = Position.create name;
  }
